# -*- coding: utf8 -*-

import cPickle
import json
import logging
import os
import math
import re
import urllib2

from bs4 import BeautifulSoup

URL = "http://www.atpworldtour.com/Share/Event-Draws.aspx?"\
      "Year=2012&EventId=%d&Draw=%s"

MEN_LS = "http://www.atpworldtour.com/Handlers/EventList.aspx?state=f"
WOMEN_LS = "http://www.wtatennis.com/page/Scores/"

# info about the tournaments... city, base singles, base doubles
WTA_TOURNAMENTS = {
    'MUTUA MADRID OPEN': ("Madrid", 64, 32),
    "INTERNAZIONALI BNL D'ITALIA": ("Roma", 64, 32),
}

ROUNDS_NAME = {
    'Quarterfinals': 8,
    'Semifinals': 4,
    'Final': 2,
}

ROUND = {
    1: 'W',
    2: 'F',
    4: 'S',
    8: 'C',
    16: 'O',
}


# log setup
_rootlogger = logging.getLogger("")
_rootlogger.setLevel(logging.INFO)
_handler = logging.StreamHandler()
_handler.setLevel(logging.INFO)
#>>> formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
#>>> ch.setFormatter(formatter)
_rootlogger.addHandler(_handler)


class PlayerInfo(object):
    _fname = 'players.db'

    def __init__(self):
        self.log = logging.getLogger("PlayerInfo").info

        if os.path.exists(self._fname):
            with open(self._fname, 'rb') as fh:
                self.db = cPickle.load(fh)
        else:
            self.db = {}
        self.log("DB ready: %d items", len(self.db))

    def save(self):
        """Save current data."""
        with open(self._fname, 'wb') as fh:
            cPickle.dump(self.db, fh)

    def get(self, name, url):
        """Return the player info.

        If the info is not there, get it from the web.
        """
        try:
            data = self.db[name]
        except:
            self.db[name] = data = self._web_get(url)
            self.save()
        return data

    def _web_get(self, url):
        """Get player info from the web."""
        self.log("Getting data from web: %r", url)
        response = urllib2.urlopen(url).read()
        soup = BeautifulSoup(response)
        nat = soup.find('div', id='playerBioInfoFlag')
        nat = nat.text.strip()
        self.log("Nationality: %r", nat)
        return nat


class Player(object):
    db = PlayerInfo()

    def __init__(self, data1, data2=None):
        name1, url1 = data1
        if data2 is None:
            self.name = name1.split(",")[0]
            self.nations = [self.db.get(name1, url1)]

            # get the short name
            parts = name1.split(", ")
            self.short_name = u"%s. %s" % (parts[1][0], parts[0])
        else:
            name2, url2 = data2
            self.name = "%s/%s" % (name1.split(",")[0], name2.split(",")[0])
            self.nations = [self.db.get(name1, url1), self.db.get(name2, url2)]

            # get the short name
            parts1 = name1.split(", ")
            parts2 = name2.split(", ")
            self.short_name = u"%s. %s / %s. %s" % (
                            parts1[1][0], parts1[0], parts2[1][0], parts2[0])

    def check_nation(self, nation):
        """Return if any of the player(s) is from that nation."""
        return nation in self.nations

    def __str__(self):
        return u"<P: %s>" % (self.name,)
    __repr__ = __str__


class Tournament(object):
    def __init__(self, tid, ttype):
        self.playing = {}
        self.sequence = {}
        self.log = logging.getLogger("Tournament").info
        self.player_conversion = {}
        self.ttype = ttype

        url = URL % (tid, ttype)
        response = urllib2.urlopen(url).read()
        self.soup = BeautifulSoup(response)
        self.city = self.soup.find('p', 'tournamentSubTitle').text.split(',')[0]
        self.log("Init, tournament: %s (%d)", self.city, tid)

    def _get_t_round(self, n_round, len_players):
        try:
            t_round = ROUND[len_players]
        except KeyError:
            t_round = str(n_round)
        return t_round

    def calculate(self):
        n_round = 0
        while True:
            n_round += 1
            col = self.soup.find('td', 'col_%d' % (n_round,))
            if n_round == 1:
                players = col.find_all('div', 'player')
                self._first_round(self._grouper(players), n_round, len(players))
            else:
                players = col.find_all('p', 'player')
                self._other_rounds(self._grouper(players), n_round, len(players))

            if len(players) == 1:
                self.log("Finishing because tournament done.")
                break
            if not self.playing:
                self.log("Finishing because no more players.")
                break

        player_info = [(p.name, s) for p, s in self.sequence.items()]
        return self.city, player_info

    def _grouper(self, players):
        """Group the players by 1 or 2, according to the tournament type."""
        players = iter(players)
        if self.ttype[-1] == 'd':
            # doubles
            while True:
                try:
                    p1 = players.next()
                    p2 = players.next()
                except StopIteration:
                    break

                if p1.a is None:
                    assert p2.a is None
                    yield None
                else:
                    yield ((p1.a.text, p1.a['href']), (p2.a.text, p2.a['href']))
        else:
            # singles
            for p in players:
                if p.a is None:
                    yield None
                else:
                    yield ((p.a.text, p.a['href']),)

    def _first_round(self, players, n_round, len_players):
        t_round = self._get_t_round(n_round, len_players)
        print "\n---"
        print t_round
        posics = {} # double entry
        for pos, player_info in enumerate(players):
            if player_info is not None:
                player = Player(*player_info)

                posics[pos] = player
                posics[player] = pos
                if player.check_nation('Argentina'):
                    self.playing[(2, pos // 2)] = player
                    self.sequence[player] = [t_round]
                self.player_conversion[player.short_name] = player
            else:
                posics[pos] = None

        # mark the preclassifieds
        for name, rnd in self.sequence.items():
            pos = posics[name]
            a, b = divmod(pos, 2)
            opposite = a * 2 + int(not b)
            if posics[opposite] is None:
                self.sequence[name] = ['P']

    def _other_rounds(self, players, n_round, len_players):
        t_round = self._get_t_round(n_round, len_players)
        print t_round
        for pos, player_info in enumerate(players):
            key = (n_round, pos)
            try:
                prev_player = self.playing.pop(key)
            except KeyError:
                # don't care, not tracking any user
                continue

            if player_info is None:
                # winner still didn't decided
                self.sequence[prev_player].append('?')
                continue

            short_name = ' / '.join(re.sub(' \[\d\]', '', x[0])
                                    for x in player_info)
            player = self.player_conversion[short_name]
            print " ", player

            if player is prev_player:
                # tracked user won!
                self.playing[(n_round + 1, pos // 2)] = prev_player
                self.sequence[prev_player].append(t_round)
            else:
                # lost, :(
                self.sequence[prev_player].append('|')


class LiveScores(object):
    """Handle live scores info."""
    _fname = 'livescores.db'

    def __init__(self):
        self.log = logging.getLogger("LiveScores").info
        if os.path.exists(self._fname):
            with open(self._fname, 'rb') as fh:
                self.db = cPickle.load(fh)
        else:
            self.db = {}
        self.log("Init, live scores, db items: %d", len(self.db))

    def save(self):
        """Save current data."""
        with open(self._fname, 'wb') as fh:
            cPickle.dump(self.db, fh)

    def _get_base_round(self, tour, match):
        """Get the base quantity of the tour according to singles/doubles."""
        # get base without preclassified
        if match['team'][0]['p2ln']:
            # doubles
            base_noprec = tour['ddrwsz']
        else:
            # singles
            base_noprec = tour['sdrwsz']

        # get real base round
        exp = math.log(base_noprec) / math.log(2)
        base_round = int(2 ** round(exp))
        assert base_round >= base_noprec
        return base_round

    def get(self, maxlines=None, dry=False):
        """Return info for all tournaments."""
        matches = []
        for match in self._get_matches():
            matches.append(match)
            if maxlines is not None and len(matches) >= maxlines:
                self.log("Stopping search because max hit")
                break

        self.log("Total new matches: %d", len(matches))
        if not dry:
            self.save()
        return matches

    def _get_matches(self):
        """Return info for all tournaments."""
        matches = []

        # atp data
        data = []
        raw = urllib2.urlopen(MEN_LS).read()
        self.log("Men web page read ok")
        data = json.loads(raw)
        self.log("Men JSON converted ok")

        for tour in data:
            try:
                for match in self._get_atp_tourninfo(tour):
                    yield match
            except Exception, err:
                self.log("BROKEN (%s): %r", err, tour)
                raise

        # wta data
        raw = urllib2.urlopen(WOMEN_LS).read()
        self.log("Women web page read ok")
        data = self._proc_wta(raw)

        for tour_name, info in data.items():
            try:
                for match in self._get_wta_tourninfo(tour_name, info):
                    yield match
            except Exception, err:
                self.log("BROKEN (%s): %r", err, info)
                raise

    def _proc_wta(self, raw):
        """Process WTA data, as it's not simple."""
        soup = BeautifulSoup(raw)
        article = soup.find('div', 'articleBody')
        self.log("Women JSON converted ok")

        def sanitize(t):
            """Remove html crap."""
            t = re.sub("<.*?>", "", t)
            t = re.sub("^\(.*?\) ", "", t)
            return t.strip()

        tours = {}
        partes = unicode(article).split("<br/>")
        team_mode_single = None
        active_round = None
        for p in partes:
            if not p:
                continue
            if "www.wtatennis.com/page/Tournaments/Info" in p:
                tour_key = sanitize(p)
                tour_name, base_single, base_double = WTA_TOURNAMENTS[tour_key]
                tours[tour_name] = results = []
                continue

            m = re.match("(\S*) - (\S*) Round", p)
            if m:
                s_mode, s_rnd = map(sanitize, m.groups())
                team_mode_single = {'Singles': True, 'Doubles': False}[s_mode]
                exp = {'First': 0, 'Second': 1, 'Third': 2}[s_rnd]
                if team_mode_single:
                    active_base = base_single
                else:
                    active_base = base_double
                active_round = active_base / 2 ** exp
                continue

            m = re.match("(\S*) - (\S*)", p)
            if m:
                s_mode, s_rnd = map(sanitize, m.groups())
                team_mode_single = {'Singles': True, 'Doubles': False}[s_mode]
                if team_mode_single:
                    active_base = base_single
                else:
                    active_base = base_double
#                if s_rnd == 'First':
#                    active_round = active_base
#                else:
                active_round = ROUNDS_NAME[s_rnd]
                continue

            m = re.match("(.*?)(\(\S{3,7}\)) d\. (.*?)(\(\S{3,7}\))(.*)", p)
            if m:
                elems = m.groups()
                elems = map(sanitize, elems)
                results.append((active_base, active_round, elems))

        return tours

    def _filter_matches(self, tour_data, rnd, team1, team2):
        """Get the info, filter the matches and prepare new info if any."""
        if u"ARG" in team1.nation:
            interest = team1
            other = team2
            won = True
        elif u"ARG" in team2.nation:
            interest = team2
            other = team1
            won = False
        else:
            # don't care
            return None

        data_key = (rnd, team1.name, team2.name)
        if data_key in tour_data:
            # old
            return None

        tour_data.add(data_key)
        return (interest, other, won)

    def _get_wta_tourninfo(self, location, info):
        """Return info for specific tournament, for women."""
        tour_data = self.db.setdefault('wta-%s' % (location,), set())
        self.log("Getting tour info for %r (len=%d)", location, len(info))

        for (base_round, rnd, match) in info:
            name1, nation1, name2, nation2, scores = match
            team1 = Team.from_wta(name1, nation1, scores, first=True)
            team2 = Team.from_wta(name2, nation2, scores, first=False)

            result = self._filter_matches(tour_data, rnd, team1, team2)
            if result is not None:
                interest, other, won = result
                new_info = (location, base_round, rnd, interest, other, won)
                self.log("New data found: %r", new_info)
                yield new_info

    def _get_atp_tourninfo(self, tour):
        """Return info for specific tournament, for men."""
        tour_data = self.db.setdefault(tour['eid'], set())
        info = tour['mlst']
        location = tour['loc'].split(",")[0]
        self.log("Getting tour info for %r (len=%d)", location, len(info))

        for match in info:
            base_round = self._get_base_round(tour, match)
            team1, team2 = map(Team.from_atp, match['team'])
            if match['rnd'] == 'QRF':
                # qualys, and still don't know how to read this
                continue
            try:
                rnd = ROUNDS_NAME[match['rnd']]
            except KeyError:
                # format: R64, R16, etc
                rnd = int(match['rnd'][1:])

            if match['win'] == 2:
                # the other way around
                team1, team2 = team2, team1
            result = self._filter_matches(tour_data, rnd, team1, team2)
            if result is not None:
                interest, other, won = result
                new_info = (location, base_round, rnd, interest, other, won)
                self.log("New data found: %r", new_info)
                yield new_info


class Team(object):
    """Team for the live scores."""
    def __init__(self, name, nation, singles, scores, tbreaks, retired=False):
        self.__dict__ = locals()

    @classmethod
    def from_wta(cls, name, nation, scores, first):
        nation = re.sub("[()]", "", nation)
        if '/' in nation:
            nation = nation.split('/')
            singles = False
        else:
            nation = [nation]
            singles = True

        real_tbreaks = [0] * 5
        real_scores = [0] * 5
        if first:
            idx = 0
        else:
            idx = 1

        retired = False
        for i, sc in enumerate(scores.split()):
            if sc == 'ret.':
                if not first:
                    # if it says 'ret.', obviously the retired is the one
                    # who lost the match
                    retired = True
                break

            m = re.match("(\d+)\((\d+)\)", sc)
            if m:
                normal_scores, tbreak_scores = m.groups()
            else:
                normal_scores = sc
                tbreak_scores = None

            if len(normal_scores) == 2:
                # normal score
                splitted = normal_scores[0], normal_scores[1]
            else:
                # bigger numbers
                splitted = normal_scores[:2], normal_scores[2:]
            real_scores[i] = int(splitted[idx])

            if tbreak_scores:
                tb_loser = int(tbreak_scores[0])
                tb_winner = max(7, tb_loser + 2)
                if splitted[0] > splitted[1]:
                    expanded_tb = (tb_winner, tb_loser)
                else:
                    expanded_tb = (tb_loser, tb_winner)
                real_tbreaks[i] = expanded_tb[idx]

        return cls(name, nation, singles, real_scores, real_tbreaks, retired)

    @classmethod
    def from_atp(cls, team):
        # name and nation
        if team['p2ln']:
            # doubles
            name = u"%s/%s" % (team['p1ln'], team['p2ln'])
            nation = [team['p1cnty'], team['p2cnty']]
            singles = False
        else:
            # singles
            name = u"%s %s" % (team['p1fn'], team['p1ln'])
            nation = [team['p1cnty']]
            singles = True

        # scores
        scores = [team.get("s%d" % i, 0) for i in range(1,6)]
        tbreaks = [team.get("t%d" % i, 0) for i in range(1,6)]
        return cls(name, nation, singles, scores, tbreaks)

    def __unicode__(self):
        return u"<T: %s (%s)>" % (self.name, self.nation)

    def __repr__(self):
        return repr(self.__unicode__())


if __name__ == '__main__':
    ls = LiveScores()
    print ls.get(2)
#    for tid in (1536,):
#        print "\n\n========== SSS"
#        t = Tournament(tid, 'ms')
#        city, data = t.calculate()
#        print "\n%s:" % (city,)
#        for player, seq in data:
#            print " ", player, '-'.join(seq)
#
#        print "\n\n========== DDD"
#        t = Tournament(tid, 'md')
#        city, data = t.calculate()
#        print "\n%s:" % (city,)
#        for player, seq in data:
#            print " ", player, '-'.join(seq)
